#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Dec  4 11:23:33 CET 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: stats.py -*-
# -*- purpose: -*-

'''
A class that performs Gaussian statistical approximation
and keep the information in the object.

'''

from math import sqrt, exp, pi

from contextlib import suppress
import warnings

import pyhisto
from pyhisto import Histogram

# In the next part, I try to make sure the module can be
# imported even if matplotlib is not installed.
try:
    import matplotlib.pyplot as plt
except ImportError:
    class Ghost:
        ''' ghost object just to fill'''
        def __getattribute__(self, name):
            def nothing(*w, **kw):
                warnings.warn('Calling a plotting function while `matplotlib` is not available')
            return nothing
    plt = Ghost()


class GaussFit:
    '''From an histogram, extract Gaussian 'fitted' parameters '''
    def __init__(self,
                 h: Histogram):
        '''Computes sum, average x, stddev, max, bin'''
        h1 = h
        bg = h1.empty_copy()
        bg_l = h1[0].count
        bg_r = h1[-1].count
        for i, b in enumerate(bg):
            b.count = bg_l + (i / float(len(bg))) * (bg_r - bg_l)
        # making bg substracted histogram
        hsbgsub = h1.copy()
        hsbgsub -= bg
        for b in hsbgsub:
            if b.count < 0:
                b.count = 0
        # Now, counting the sum
        S = sum(hsbgsub) * hsbgsub[0].width()
        # Mean value
        with suppress(ZeroDivisionError):
            M = sum((b.center() * b.count for b in hsbgsub)) / sum(hsbgsub)
        # Stddev
        with suppress(ZeroDivisionError):
            Stdev = sum((b.center() ** 2. * b.count for b in hsbgsub)) / sum(hsbgsub)
        Stdev -= M * M
        Stdev = sqrt(Stdev)
        # making the gaussian, and residual
        G = hsbgsub.empty_copy()
        R = G.copy()
        for i, b in enumerate(G):
            b.count = S / sqrt(2. * pi * Stdev * Stdev) * exp(-(b.center() - M) ** 2. / 2. / Stdev / Stdev)
            R[i].count = hsbgsub[i] - b.count
        self.h1 = h1
        self.background = bg
        self.bgsubstracted = hsbgsub
        self.gaussian = G
        self.residuals = R
        self.integral = S
        self.mean = M
        self.stdev = Stdev

    def add_to_plot(self):
        '''If a figure is already 'in progress',  add to the plot'''
        plt.step(*self.h1.xy())
        plt.step(*self.background.xy())
        plt.step(*self.bgsubstracted.xy())
        plt.step(*self.gaussian.xy())
        plt.step(*self.residuals.xy())

    def plot(self,
             target='screen'):
        '''Plot in a new figure (alone) and send to `target` (screen or file)'''
        plt.figure()
        plt.grid(True)
        plt.step(*self.h1.xy(), label='original')
        plt.step(*self.background.xy(), label='background')
        plt.step(*self.bgsubstracted.xy(), label='bacground subracted')
        plt.step(*self.gaussian.xy(), label='Gaussian')
        plt.step(*self.residuals.xy(), label='residual')
        plt.legend()
        if target == 'screen':
            plt.show()
        else:
            plt.savefig(target)

    def __repr__(self):
        return str({'integral': self.integral,
                    'mean': self.mean,
                    'stdev': self.stdev})

# end of class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Thu May  4 15:40:35 CEST 2017 -*-
# -*- copyright: GH/IPHC 2017 -*-
# -*- file: histogram.py -*-

'''
Module for handling 1D histogram

An histogram is an ordered list of bins
'''

import sys

from typing import Union, Optional

from pyhisto.bin import Bin


class Histogram:
    '''This class contains a list of 1D bins and provide interfaces to use them

    The Histogram has bins covering [xmin:xmax]
    '''

    def __init__(self,
                 nbins: int = 1,
                 xmin: float = -0.5,
                 xmax: Optional[Union[float, None]] = None,
                 fromstring=None,
                 frombins=None,
                 fromvalues=None,
                 fromdict=None,
                 fromfile=None):
        '''Initialize the histogram

        - Using a string (obtained from str) using
        ``Histogram(fromstring=z)``

        - Using a filename
        ``Histogram(fromfile=filepath)``

        - Using a list of bins (as in bins[x:y]) using
        ```Histogram(frombins=l)``

        - Using a dict (obtained from repr)
        ``Histogram(fromdicy=d)``

        - Using a list of values
        ``Histogram(fromvalues=l)``

        - Using the 'natural way'
        ``Histogram(nbins, xmin, xmax)``
        '''
        if fromstring:
            self._fromstring(fromstring)
        elif fromfile:
            self._fromfile(fromfile)
        elif frombins:
            self._frombins(frombins)
        elif fromdict:
            self._fromdict(fromdict)
        elif fromvalues:
            self._fromvalues(fromvalues)
        else:
            self.nbins = nbins
            self.xmin = xmin
            self.xmax = xmax
            if self.xmax is None:
                self.xmax = xmin + nbins
            self.xmin, self.xmax = min(self.xmin, self.xmax), max(self.xmin, self.xmax)
            self.binwidth = (self.xmax - self.xmin) / float(self.nbins)
            self.bins = [Bin(self.xmin + i * self.binwidth,
                             self.binwidth) for i in range(self.nbins)]
            self.outofrange = 0

    def __del__(self):
        del self.bins
        del self.nbins
        del self.xmax
        del self.xmin
        del self.outofrange
        del self.binwidth

    def __sizeof__(self):
        return sum([sys.getsizeof(self.bins),
                    sys.getsizeof(self.nbins),
                    sys.getsizeof(self.xmin),
                    sys.getsizeof(self.xmax),
                    sys.getsizeof(self.binwidth),
                    sys.getsizeof(self.outofrange),
                    0])

    def _repr_(self):
        ''' lite alternative to the real __repr__'''
        return "<histogram1D nbins={0} xmin={1} xmax={2}>".format(self.nbins,
                                                                  self.xmin,
                                                                  self.xmax)

    def __repr__(self):
        '''Represent the histogram as a dictionnary, list'''
        z = ''
        z += "{{'nbins': {0}, 'xmin': {1}, 'xmax': {2}, 'oor': {3}, \n".format(self.nbins,
                                                                               self.xmin,
                                                                               self.xmax,
                                                                               self.outofrange)
        z += "'bins': [ \n"
        for b in self.bins:
            z += "\t {0}, \n".format(repr(b))
        z += "        ]\n}"
        return z

    def __str__(self):
        '''Dump the histogram in string format'''
        z = ''
        z += '# nbins = {0}\n'.format(self.nbins)
        z += '# xmin = {0}\n'.format(self.xmin)
        z += '# xmax = {0}\n'.format(self.xmax)
        for i in range(0, self.nbins):
            z += "{0} \n".format(str(self.bins[i]))
        z += '# out of range = {0}\n'.format(self.outofrange)
        return z

    def _fromfile(self, fp):
        '''Creates the histogram from a filepath'''
        self._fromstring(open(fp, 'r').read())

    def _fromstring(self, instr):
        '''This function imports an histogram from a string,
        formatted as the output of __str__'''
        lines = instr.split('\n')
        self.xmax = None
        self.xmin = 0
        self.binwidth = 1.
        self.nbins = None
        self.bins = []
        self.outofrange = 0
        for l in lines:
            if l.startswith("# nbins"):
                self.nbins = int(l.split('=')[1])
                self.bins = []
            elif l.startswith("# xmin"):
                self.xmin = float(l.split('=')[1])
            elif l.startswith("# xmax"):
                self.xmax = float(l.split('=')[1])
            elif l.startswith("# out of range"):
                self.outofrange = float(l.split('=')[1])
            elif l.startswith('#'):
                pass
            elif len(l) <= 1:
                pass
            else:
                try:
                    self.bins.append(Bin(fromstring=l))
                except Exception:
                    pass
        if not self.nbins:
            self.nbins = len(self.bins)
        if not self.xmax:
            self.xmax = self.bins[-1].upedge
        self.binwidth = (self.xmax - self.xmin) / float(self.nbins)

    def _frombins(self, bs):
        ''' This function imports an hisogram from a list of bins '''
        self.bins = sorted(bs, key=lambda x: x.lowedge)
        self.xmin = self.bins[0].lowedge
        self.xmax = self.bins[-1].upedge
        self.nbins = len(bs)
        self.binwidth = self.bins[0].width()
        self.outofrange = 0

    def _fromvalues(self, lsv):
        ''' import from a list of values '''
        self.nbins = len(lsv)
        self.xmin = -0.5
        self.bins = []
        for i in range(self.nbins):
            self.bins.append(Bin(lowedge=i - 0.5, width=1, count=lsv[i]))
        self.xmax = self.bins[-1].upedge
        self.binwidth = self.bins[0].width()
        self.outofrange = 0

    def _fromdict(self, json):
        '''create the histogram from a dictionnary'''
        raise NotImplementedError

    def dict(self):
        '''returns a dictionnary representation'''
        return {
            'nbins': self.nbins,
            'xmin': self.xmin, 'xmax': self.xmax,
            'oor': self.outofrange,
            'bins': [vars(b) for b in self.bins],
        }

    # Next: setting a Container like behavior
    def __getitem__(self, k):
        return self.bins[k]

    def __setitem__(self, k, v):
        self.bins[k] = v

    def __len__(self):
        return self.nbins

    def index(self, x):
        ''' Return the index corresponding to x (slow version)'''
        for i in range(len(self.bins)):
            if x in self.bins[i]:
                return i
        raise IndexError

    def find(self, x):
        ''' Return the bin corresponding to x (slow)'''
        for b in self.bins:
            if x in b:
                return b
        raise IndexError

    # fast index uses xmin, width to go fast
    def fast_index(self, x):
        ''' find the index fast '''
        return int((x - self.xmin) / self.binwidth)

    def slice(self, _from=None, to=None):
        '''Returns a sub histogram'''
        # default from is first bin [:ito]
        ifrom = 0
        if _from:
            ifrom = self.index(_from)
        # default to is last bin [ifrom:]
        ito = self.nbins-1
        if to:
            ito = self.index(to)
        # return bins slice
        return Histogram(frombins=self.bins[ifrom:ito + 1])

    def __iter__(self):
        ''' provides iteration over bin directly '''
        return iter(self.bins)

    def empty(self, who='all'):
        ''' Empty the counts from all bins '''
        if who == 'all':
            self.outofrange = 0
            for b in self.bins:
                b.count = 0
        elif isinstance(who, list):
            for x in who:
                self.find(x).count = 0
        else:
            self.find(who).count = 0

    def copy(self):
        ''' Return a copy of the histogram '''
        return Histogram(fromstring=str(self))

    def empty_copy(self):
        ''' Return a new Histogram with the same xmax, xmin, ... but all counts to 0'''
        hout = self.copy()
        hout.empty()
        return hout

    def fill(self, x, w=1):
        '''Filling the bin at `x` with `w`'''
        try:
            self.find(x).fill(w)
        except IndexError:
            self.outofrange += w

    def fast_fill(self, x, w=1):
        '''fast verion of filling'''
        try:
            self.bins[int((x - self.xmin) / self.binwidth)].fill(w)
        except IndexError:
            self.outofrange += w

    def xy(self, x_pos='low'):
        ''' returns the bin postions and counts as tuple to be used in plotting with pyplot'''
        return zip(*[b.xy(x_pos) for b in self.bins])

    # scalar operations: add, sub, ul, divide
    def __iadd__(self, o):
        ''' Increment all bins by o if float or bin by bin if histogram '''
        if isinstance(o, Histogram):
            for b1, b2 in zip(self.bins, o.bins):
                b1 += b2
        else:
            for b in self.bins:
                b += float(o)
        return self

    def __add__(self, o):
        ''' Return a new histogram with counts added '''
        newH = self.empty_copy()
        if isinstance(o, Histogram):
            for a, b1, b2 in zip(newH, self.bins, o.bins):
                a.count = b1.count+b2.count
        else:
            for a, b in zip(newH, self.bins):
                a.count = b.count + float(o)
        return newH

    def __radd__(self, o):
        ''' This type of operation makes no sense -> TypeError '''
        raise TypeError('Adding an histogram to something other than an histogram makes no sense')

    def __isub__(self, o):
        ''' Decrement all bins by o if float or bin by bin if histogram '''
        if isinstance(o, Histogram):
            for b1, b2 in zip(self.bins, o.bins):
                b1 -= b2
        else:
            for b in self.bins:
                b -= float(o)
        return self

    def __sub__(self, o):
        ''' Return a new histogram with counts substracted '''
        newH = self.empty_copy()
        if isinstance(o, Histogram):
            for a, b1, b2 in zip(newH, self.bins, o.bins):
                a.count = b1.count - b2.count
        else:
            for a, b in zip(newH, self.bins):
                a.count = b.count - float(o)
        return newH

    def __rsub__(self, o):
        ''' This type of operation makes no sense -> TypeError '''
        raise TypeError('Adding an histogram to something other than an histogram makes no sense')

    def __imul__(self, o):
        ''' Multiply all bins by o if float or bin by bin if histogram '''
        if isinstance(o, Histogram):
            for b1, b2 in zip(self.bins, o.bins):
                b1 *= b2
        else:
            for b in self.bins:
                b *= float(o)
        return self

    def __mul__(self, o):
        ''' Return a new histogram with counts multiplied '''
        newH = self.empty_copy()
        if isinstance(o, Histogram):
            for a, b1, b2 in zip(newH, self.bins, o.bins):
                a.count = b1.count * b2.count
        else:
            for a, b in zip(newH, self.bins):
                a.count = b.count * float(o)
        return newH

    def __rmul__(self, o):
        ''' This type of operation makes no sense -> TypeError '''
        raise TypeError('Adding an histogram to something other than an histogram makes no sense')

    def __idiv__(self, o):
        ''' Divide all bins by o if float or bin by bin if histogram '''
        if isinstance(o, Histogram):
            for b1, b2 in zip(self.bins, o.bins):
                b1 /= b2
        else:
            for b in self.bins:
                b /= float(o)
        return self

    def __div__(self, o):
        ''' Return a new histogram with counts divided '''
        newH = self.empty_copy()
        if isinstance(o, Histogram):
            for a, b1, b2 in zip(newH, self.bins, o.bins):
                a.count = b1.count / b2.count
        else:
            for a, b in zip(newH, self.bins):
                a.count = b.count / float(o)
        return newH

    def __rdiv__(self, o):
        ''' This type of operation makes no sense -> TypeError '''
        raise TypeError('Adding an histogram to something other than an histogram makes no sense')

    def autocrop(self,
                 mincount=0.0,
                 minwidth=1.0e-10):
        ''' Clean this histogram (returns self):
        - Remove bins on the side with less than mincount
        - Remove bins with width less than minwidth
        - Reorder bins according to xmin
        '''
        # remove null-width
        for b in self.bins:
            if b.width() < minwidth:
                self.bins.remove(b)
        # end for
        # scan up
        imin = 0
        for i in range(len(self.bins)):
            imin = i
            if self.bins[i].count > mincount:
                break
        # end for
        # scan down
        imax = self.nbins
        for i in reversed(range(len(self.bins))):
            imax = i + 1
            if self.bins[i].count > mincount:
                break
        # end for
        # send to _frombins to make a proper histogram
        self._frombins(self.bins[imin:imax])
        return self

    def eval(self, x, method=None):
        '''Evaluate the value of the histogram,after performing some smoothing
        so far, method is useless
        '''
        raise NotImplementedError



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Oct  9 11:31:54 CEST 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: lazy_histogram.py -*-
# -*- purpose: -*-

'''
The LazyHistogram is to be used when speed and memory are keys
(typically filling).
It is backed by the `array` module to store the values (integter)
with smaller footprint.
'''

import array

from typing import Union, Optional, List

class LazyHistogram:
    '''
    This Histogram is a liteweight version of 1D histo, for purpose of filling
    '''
    def __init__(self,
                 nbins: int = 1,
                 xmin: float = -0.5,
                 xmax: Union[float, None] = None,
                 arraytype: str = 'L'):
        self.nbins: int = nbins
        self.xmin: float = xmin
        self.xmax: float = xmin + nbins
        if not (xmax is None):
            self.xmax = xmax
        self.xmin, self.xmax = min(self.xmin, self.xmax), max(self.xmin, self.xmax)
        self.binwidth: float = (self.xmax - self.xmin) / self.nbins
        self.bins = array.array(arraytype,
								[0] * self.nbins)
        self.outofrange = 0
    # end __init__

    def __str__(self) -> str:
        '''Dump the histogram in string format'''
        z = ''
        z += '# nbins = {0}\n'.format(self.nbins)
        z += '# xmin = {0}\n'.format(self.xmin)
        z += '# xmax = {0}\n'.format(self.xmax)
        for i in range(0, self.nbins):
            lowedge = self.xmin + i * self.binwidth
            z += f"{lowedge} {self.binwidth} {self.bins[i]}\n"
        # end for bins
        z += f'# out of range = {self.outofrange}\n'
        return z
    # end __str__

    def __del__(self):
        del self.nbins
        del self.xmin
        del self.xmax
        del self.binwidth
        del self.outofrange
        del self.bins
    # end __del__

    # following: the two working functions
    def fast_index(self, x: float) -> int:
        ''' returns the index for `x` '''
        return int((x - self.xmin) / self.binwidth)

    def fast_fill(self,
                  x: float,
                  w=1):
        ''' fast filling of the histogram
        find the bin corresponding to `x` and adding `w` into its count
        '''
        try:
            self.bins[int((x - self.xmin) / self.binwidth)] += w
        except IndexError:
            self.outofrange += w

    # The following 4 fucntions provide container like behavior
    def __getitem__(self, k):
        return self.bins[k]

    def __setitem__(self, k, v):
        self.bins[k] = v

    def __len__(self) -> int:
        return len(self.bins)

    def __iter__(self):
        ''' provides iteration over bin directly '''
        return iter(self.bins)

# end of file



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Dec  4 11:23:33 CET 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: stats.py -*-
# -*- purpose: -*-

'''
Performs a real fit of a gaussian function using scipy methods!
'''

from math import sqrt, exp, pi
import warnings

from pyhisto import Histogram

from pyhisto.tools import GaussFit

# Here, in case the requirements are not met...
class Ghost:
    ''' ghost object just to fill'''
    def __getattribute__(self, name):
        def nothing(*w, **kw):
            warnings.warn('Calling a plotting function while `matplotlib` is not available')
        return nothing


try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = Ghost()

try:
    import numpy as np
    import scipy
    from scipy.optimize import curve_fit
except ImportError:
    raise ImportError("Could not implement Real Fit because `numpy` and `scipy` are not availble")


class GaussRealFit:
    '''From an histogram, extract Gaussian 'fitted' parameters '''
    def __init__(self,
                 h: Histogram):
        '''computes sum, average x, stddev, max, bin'''
        self.h1 = h
        _simplefit = GaussFit(self.h1)

        def GaussFunc(x,
                      S, M, R,
                      B, SL):
            return B + SL * (x - M) + S / np.sqrt(2. * pi * R * R) * np.exp(-(x - M) ** 2. / 2. / R / R)

        first_guess = [_simplefit.integral,
                       _simplefit.mean,
                       _simplefit.stdev,
                       self.h1[0].count,
                       0.0]
        limits = list(zip((0., np.inf),
                          (self.h1[0].lowedge, self.h1[-1].upedge),
                          (0., np.inf),
                          (-np.inf, np.inf),
                          (-np.inf, np.inf)
                          ))
        fitrslt, cov = curve_fit(GaussFunc,
                                 *self.h1.xy(),
                                 p0=first_guess,
                                 bounds=limits)
        uncerts = np.sqrt(np.diag(cov))
        self.integral = fitrslt[0]
        self.integral_u = uncerts[0]
        self.mean = fitrslt[1]
        self.mean_u = uncerts[1]
        self.stdev = fitrslt[2]
        self.stdev_u = uncerts[2]
        # Filling histograms for display
        self.background = self.h1.empty_copy()
        for b in self.background:
            b.count = fitrslt[3] + fitrslt[4] * (b.center() - self.mean)
        self.bgsubstracted = self.h1.copy()
        self.bgsubstracted -= self.background
        self.gaussian = self.h1.empty_copy()
        for b in self.gaussian:
            b.count = self.integral / sqrt(2. * pi * self.stdev * self.stdev) * exp(-(b.center() - self.mean) ** 2. / 2. / self.stdev / self.stdev)
        self.residuals = self.bgsubstracted.copy()
        self.residuals -= self.gaussian

    def add_to_plot(self):
        '''If a figure is already 'in progress',  add to the plot'''
        plt.step(*self.h1.xy())
        plt.step(*self.background.xy())
        plt.step(*self.bgsubstracted.xy())
        plt.step(*self.gaussian.xy())
        plt.step(*self.residuals.xy())

    def plot(self,
             target='screen'):
        '''Plot in a new figure (alone) and send to `target` (screen or file)'''
        plt.figure()
        plt.grid(True)
        plt.step(*self.h1.xy(), label='original')
        plt.step(*self.background.xy(), label='background')
        plt.step(*self.bgsubstracted.xy(), label='bacground subracted')
        plt.step(*self.gaussian.xy(), label='Gaussian')
        plt.step(*self.residuals.xy(), label='residual')
        plt.legend()
        if target == 'screen':
            plt.show()
        else:
            plt.savefig(target)

    def __repr__(self) -> str:
        return str({'integral': self.integral,
                    'integral_u': self.integral_u,
                    'mean': self.mean,
                    'mean_u': self.mean_u,
                    'stdev': self.stdev,
                    'stdev_u': self.stdev_u})
# end of class

# EOF



#!/usr/env python
'''
Tools for histogram
'''

from .parzen_estimator import ParzenEstimator
from .stats import Stats
from .findpeaks import FindPeaks
from .gaussfit import GaussFit
from .gaussrealfit import GaussRealFit



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: 2019 05 02 -*-
# -*- copyright: GH/IPHC 2019 -*-
# -*- file: findpeaks.py -*-
# -*- purpose: -*-

'''
FindPeaks method
'''
from pyhisto import Histogram


class FindPeaks:
    '''From an histogram object, find n peakes'''
    def __init__(self,
                 h: Histogram,
                 npeaks: int = 1,
                 width: float = 1.0):
        '''
        npeak=int, width=float (if negative: relative width)
        '''
        sorted_bins = sorted(h.bins, reverse=True)
        self.peaks = []
        width_func = lambda x: width
        if width < 0.:
            width_func = lambda x: abs(width) * x
        self.peaks.append(sorted_bins.pop(0))
        for b in sorted_bins:
            if all([abs(p.lowedge - b.lowedge) > width_func(p.lowedge) for p in self.peaks]) \
               and len(self.peaks) < npeaks:
                self.peaks.append(b)
            # end if
        self.peaks = sorted(self.peaks, key=lambda b: b.lowedge)
        # end for
# end findpeaks

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Fri Aug 11 10:22:50 CEST 2017 -*-
# -*- copyright: GH/IPHC 2017 -*-
# -*- file: histogram2D.py -*-
# -*- purpose: -*-

'''
The module contain a Ghost Histogram2D class.
'''


class GhostHistogram2D:
    ''' Implement 2D hisotgram'''

    def __init__(self,
                 xnbins=1, xmin=None, xmax=None,
                 ynbins=1, ymin=None, ymax=None):
        '''Initilize the Ghost Hisotgram2D

        ``Histogram2D(nxbins, xmin, xmax, nybins, ymin, ymax)``
        '''
        self.xnbins = 1
        self.ynbins = 1
        self.xmin = -0.5
        self.xmax = 0.5
        self.xbinwidth = 1
        self.ymin = -0.5
        self.ymax = 0.5
        self.ybinwidth = 1
        self.bins = [[0]]
        self.outofrange = 0
        self.export_threshold = 0.
    # end __init__

    def __str__(self):
        '''Dump the histogram in string format'''
        z = f'# xnbins = {self.xnbins}\n'
        z += f'# xmin = {self.xmin}\n'
        z += f'# xmax = {self.xmax}\n'
        z += f'# ynbins = {self.ynbins}\n'
        z += f'# ymin = {self.ymin}\n'
        z += f'# ymax = {self.ymax}\n'
        z += f'# out of range = {self.outofrange}\n'
        return z
    # end __str__

    def __repr__(self):
        '''Writting the histogram2D as a dictionnary'''
        z = "{ \n"
        z += f"   'xnbins':{self.xnbins},\n"
        z += f"   'xmin':{self.xmin},\n"
        z += f"   'xmax':{self.xmax},\n"
        z += f"   'ynbins':{self.ynbins},\n"
        z += f"   'ymin':{self.ymin},\n"
        z += f"   'ymax':{self.ymax},\n"
        z += f"   'outofrange':{self.outofrange},\n"
        z += f"   'bins':{self.bins},\n"
        z += "}"
        return z
    # end __repr__

    def fast_fill(self, x, y, w=1):
        '''Add one (or w) counts to the bin, in a fast way'''
        pass
# end class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Fri Aug 11 10:22:50 CEST 2017 -*-
# -*- copyright: GH/IPHC 2017 -*-
# -*- file: histogram2D.py -*-
# -*- purpose: -*-

'''
The module contains the Lazy Histogram2D class.
'''


import array

from pyhisto.histogram2D import Histo2DError


class LazyHistogram2D:
    ''' Implement 2D hisotgram'''

    def __init__(self,
                 xnbins=1, xmin=None, xmax=None,
                 ynbins=1, ymin=None, ymax=None,
                 arraytype='H'):
        '''Initilize the Lazy Hisotgram2D
        ``Histogram2D(nxbins, xmin, xmax, nybins, ymin, ymax)``
        '''
        self.xnbins = xnbins
        self.ynbins = ynbins
        if xmin is None:
            xmin = -0.5
        if xmax is None:
            xmax = xmin + xnbins
        if ymin is None:
            ymin = -0.5
        if ymax is None:
            ymax = ymin + ynbins
        if xmin == xmax:
            raise Histo2DError("X Bounds values in histogram should be different")
        if ymin == ymax:
            raise Histo2DError("Y Bounds values in histogram should be different")
        xmin, xmax = min(xmax, xmin), max(xmax, xmin)
        ymin, ymax = min(ymax, ymin), max(ymax, ymin)
        self.xmin = xmin
        self.xmax = xmax
        self.xbinwidth = (self.xmax - self.xmin) / float(self.xnbins)
        self.ymin = ymin
        self.ymax = ymax
        self.ybinwidth = (self.ymax - self.ymin) / float(self.ynbins)
        self.bins = array.array(arraytype, [0 for j in range(self.ynbins * self.xnbins)])
        self.outofrange = 0
        self.export_threshold = 0.
    # end __init__

    def __str__(self):
        '''Dump the histogram in string format'''
        z = ''
        z += f'# xnbins = {self.xnbins}\n'
        z += f'# xmin = {self.xmin}\n'
        z += f'# xmax = {self.xmax}\n'
        z += f'# ynbins = {self.ynbins}\n'
        z += f'# ymin = {self.ymin}\n'
        z += f'# ymax = {self.ymax}\n'
        if self.export_threshold > 0.:
            z += f'# Exporting bins with more than {self.export_threshold} counts\n'
        for i in range(0, self.xnbins):
            for j in range(0, self.ynbins):
                if self.bins[i + j * self.xnbins] > self.export_threshold:
                    z += f"{i} {j} {self.bins[i+j*self.xnbins]}\n"
        z += f'# out of range = {self.outofrange}\n'
        return z
    # end __str__

    def __repr__(self):
        '''Writting the histogram2D as a dictionnary'''
        z = "{ \n"
        z += f"   'xnbins':{self.xnbins},\n"
        z += f"   'xmin':{self.xmin},\n"
        z += f"   'xmax':{self.xmax},\n"
        z += f"   'ynbins':{self.ynbins},\n"
        z += f"   'ymin':{self.ymin},\n"
        z += f"   'ymax':{self.ymax},\n"
        z += f"   'outofrange':{self.outofrange},\n"
        z += f"   'bins':{self.bins},\n"
        z += "}"
        return z
    # end __repr__

    def fast_fill(self, x: float, y: float,
                  w: int = 1):
        '''Add one (or w) counts to the bin, in a fast way'''
        try:
            self.bins[int((x - self.xmin) / self.xbinwidth) + self.xnbins * int((y - self.ymin) / self.ybinwidth)] += w
        except IndexError:
            self.outofrange += w

# end class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Fri Feb 10 16:09:24 CET 2017 -*-
# -*- copyright: GH/IPHC 2017 -*-
# -*- file:  bin.py -*-

'''
 Module for 1-dimension bins

 A bin is used by Histograms.
 It is defined by low and up edges, a center and a count.
'''
from typing import Union, Optional

Number = Union[int, float]


class Bin:
    '''This object represent a one dimensionnal bin

    It can be represented as a string, or used as a number (equal to count).
    '''

    def __init__(self,
                 lowedge: Number = -0.5,
                 width: Number = 1,
                 count: Number = 0,
                 upedge: Optional[Union[Number, None]] = None,
                 fromstring: Optional[Union[str, None]] = None,
                 fromdict: Optional[Union[dict, None]] = None):
        ''' Create new bin

        Can be used to import from string or define a new bin:

        - To import from a string (f.ex. provided by Bin.str), use
        ``Bin(fromstring=z)```

        - Import from a dict (as after repr()): Bin(fromdict=d)

        - To define a new bin use the other key words.
        count: int, float -- The number of count (default 0)
        lowedge: float -- The lower limit of the bin (default -0.5)
        width: float -- the bin width (default 1)
        upedge: float -- the upper limit of the bin (default lowedge+width)

        The prefered way to create a bin is
        >>> Bin(count, lowedge, width)
        '''
        if fromstring:
            inputs = fromstring.split()
            if len(inputs) == 3:
                self.lowedge, width, self.count = map(float, inputs)
            elif len(inputs) == 2:
                self.lowedge = float(inputs[0]) - 0.5
                width = 1.0
                self.count = float(inputs[1])
        elif fromdict:
            self.lowedge = fromdict['from']
            self.upedge = fromdict['to']
            self.count = fromdict['count']
        else:
            self.count = count
            self.lowedge = lowedge
            self.upedge = upedge
        # end if string
        if upedge is None:
            self.upedge = self.lowedge + width
        if self.upedge < self.lowedge:
            self.lowedge, self.upedge = min(self.upedge, self.lowedge), max(self.upedge, self.lowedge)
    # end of init

    def __index__(self):
        ''' return a 'unique' indeitifer for the bin in the histogram '''
        return int("{0}{1}{2}".format(self.lowedge, self.count, self.upedge))

    def width(self):
        ''' Returns the width of the bin'''
        return self.upedge - self.lowedge

    def center(self):
        '''Returhs bin center (i.e. 0.5*(up+low))'''
        return 0.50 * (self.upedge + self.lowedge)

    def dict(self):
        ''' Returns the bin as a dictionnary '''
        return {'from': self.lowedge,
                'to': self.upedge,
                'count': self.count}

    def __repr__(self):
        return "{{'from':{0.lowedge}, 'to':{0.upedge}, 'count':{0.count}}}".format(self)

    def __str__(self):
        ''' Return as a string in a format that can be used to
        create a new bin with bin(string=)'''
        return "{0} {1} {2} ".format(self.lowedge,
                                     self.width(),
                                     self.count)

    # Type casting
    def __int__(self):
        ''' Returns count as an integer '''
        return int(self.count)

    def __float__(self):
        ''' Returns count as a float '''
        return float(self.count)

    # Addition
    def fill(self, w=1):
        ''' adds `w` to the bin count '''
        self.count += w

    def __add__(self, o):
        ''' Add (an number) to count '''
        return Bin(count=self.count + float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __radd__(self, o):
        ''' This is so that a sum(list-of-Bin) computes the sum of bin counts '''
        return self.count + float(o)

    def __iadd__(self, o):
        ''' Increment bin count '''
        self.count += float(o)
        return self

    # Substration
    def __sub__(self, o):
        ''' sub (an number) to count '''
        return Bin(count=self.count - float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __rsub__(self, o):
        '''  '''
        return Bin(count=self.count - float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __isub__(self, o):
        ''' Increment bin count '''
        self.count -= float(o)
        return self

    # Multiply
    def __mul__(self, o):
        ''' Multiply by a number '''
        return Bin(count=self.count * float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __rmul__(self, o):
        '''  '''
        return Bin(count=self.count * float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __imul__(self, o):
        ''' self multiplication '''
        self.count *= float(o)
        return self

    # Divide
    def __div__(self, o):
        ''' Multiply by a number '''
        return Bin(count=self.count / float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __rdiv__(self, o):
        '''  '''
        return Bin(count=self.count / float(o),
                   lowedge=self.lowedge,
                   upedge=self.upedge)

    def __idiv__(self, o):
        ''' self multiplication '''
        self.count /= float(o)

    # Comparison operator: we compare the counts and not the limits
    def __lt__(self, other):
        return self.count < float(other)

    def __le__(self, other):
        return self.count <= float(other)

    def __eq__(self, other):
        return self.count == float(other)

    def __ne__(self, other):
        return self.count != float(other)

    def __gt__(self, other):
        return self.count > float(other)

    def __ge__(self, other):
        return self.count >= float(other)

    def __and__(self, o):
        '''Merge the two bins (no proximity check)'''
        return Bin(count=self.count + o.count,
                   lowedge=min(self.lowedge, o.lowedge),
                   upedge=max(self.upedge, o.upedge))

    def __contains__(self, v):
        ''' Return true if v is inside Bin limits '''
        return v >= self.lowedge and v < self.upedge

    # other functions to 'sort' by bin bounadries
    def overlaps(self, other):
        ''' Determine if the bin 'other' overlaps with self's boundaries '''
        return not (other.lowedge > self.upedge
                    or other.upedge < self.lowedge)

    def higher(self, other):
        ''' determine if other bin is higher in x than self '''
        return other.lowedge >= self.upedge

    def lower(self, other):
        ''' determine if other bin is lower in x than self '''
        return other.upedge < self.lowedge

    def xy(self, x_pos='low'):
        ''' return a tuple of (x, count)

        x_pos: str -- Determine the position to return for x: low, up or center (default: low)
        '''
        y = self.count
        x = self.lowedge
        if x_pos == 'up':
            x = self.upedge
        elif x_pos == 'center':
            x = self.center()
        return (x, y)



#!/usr/env python
'''
This module provides Histogram solutions in python

It is optimized for python 3.5 and above

It includes:
- Histogram: a 1D histogram
- LazyHistogram: a simpler 1D histogram optimized for speed when filling
- GhostHistogram: same interface as a 1D histogram when filling, but does nothing
- Histogram2D: for heat maps.

'''

from .histogram import Histogram
from .lazyhistogram import LazyHistogram
from .ghosthistogram import GhostHistogram
from .histogram2D import Histogram2D
from .lazyhistogram2D import LazyHistogram2D



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Dec  4 11:23:33 CET 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: stats.py -*-
# -*- purpose: -*-

'''
Returns a stat object

(Could use a namedtuple, but kept this way for historical reason)
'''

from math import sqrt

from pyhisto import Histogram


class Stats:
    '''From an histogram object, computes statistics and keep them int he object'''
    def __init__(self,
                 h: Histogram):
        '''computes sum, average x, stddev, max, bin'''
        self.sum = sum(h)
        self.average_count = self.sum / len(h)
        self.max = max(h)
        self.mean = sum([b.center() * b.count for b in h]) / self.sum
        self.stddev = sqrt(sum([b.count * b.center() ** 2.0 for b in h]) / self.sum - self.mean ** 2.0)

    def __repr__(self):
        return f"""{{'sum': {self.sum},
'max': {self.max},
'average_count': {self.average_count},
'mean': {self.mean},
'stddev': {self.stddev},
}}"""
# end class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Fri Aug 11 10:22:50 CEST 2017 -*-
# -*- copyright: GH/IPHC 2017 -*-
# -*- file: histogram2D.py -*-
# -*- purpose: -*-

'''
The module contain the Histogram2D class.
'''


from math import floor

from typing import Union, Optional

# importing 1D histogram to export projections
from pyhisto.histogram import Histogram


class Histo2DError(Exception):
    '''2D Histogram error '''


class Histogram2D:
    ''' Implement 2D hisotgram'''
    def __init__(self,
                 xnbins: int = 1,
                 xmin: Optional[float] = None,
                 xmax: Optional[float] = None,
                 ynbins: int = 1,
                 ymin: Optional[float] = None,
                 ymax: Optional[float] = None,
                 fromstring: Optional[str] = None,
                 fromarray: Optional[Union[list, tuple]] = None,
                 fromdict: Optional[dict] = None):
        '''Initilize the Hisotgram2D

        - Using a string, as exported from str(histo2d) using
        ``Histogram2D(fromstring=z)``

        - Using a dictionnary, as obtained from repr, using
        ``Histogram2D(fromdict=d)`` (not implemented yet)

        - Using an array of values, which will be indexed naturally,
        ``Histogram2D(fromarray=a)`` (not implemented yet)

        - From 'nothing', using
        ``Histogram2D(nxbins, xmin, xmax, nybins, ymin, ymax)``
        '''
        if fromstring:
            self._fromstring(fromstring)
        elif fromarray:
            raise NotImplementedError
        elif fromdict:
            raise NotImplementedError
        else:
            self.xnbins: int = xnbins
            self.ynbins: int = ynbins
            self.xmin: float = -0.5 if xmin is None else xmin
            self.xmax: float = self.xmin + self.xnbins if xmax is None else xmax
            self.ymin: float = -0.5 if ymin is None else ymin
            self.ymax: float = self.ymin + self.ynbins if ymax is None else ymax
            if self.xmin == self.xmax:
                raise Histo2DError("X Bounds values in histogram should be different")
            if self.ymin == self.ymax:
                raise Histo2DError("Y Bounds values in histogram should be different")
            self.xmin, self.xmax = min(self.xmax, self.xmin), max(self.xmax, self.xmin)
            self.ymin, self.ymax = min(self.ymax, self.ymin), max(self.ymax, self.ymin)
            self.xbinwidth: float = (self.xmax - self.xmin) / float(self.xnbins)
            self.ybinwidth: float = (self.ymax - self.ymin) / float(self.ynbins)
            self.bins: list = [[0 for j in range(self.ynbins)]
                               for i in range(self.xnbins)]
            self.outofrange: float = 0
            self.export_threshold: float = 0.
        # end if
    # end __init__

    def _fromstring(self, instr: str):
        '''This function imports an histogram from a string,
        formatted as the output of __str__'''
        lines = instr.split('\n')
        self.outofrange = 0
        for l in lines:
            if l.startswith("# xnbins"):
                self.xnbins = int(l.split('=')[1])
            elif l.startswith("# xmin"):
                self.xmin = float(l.split('=')[1])
            elif l.startswith("# xmax"):
                self.xmax = float(l.split('=')[1])
            elif l.startswith("# ynbins"):
                self.ynbins = int(l.split('=')[1])
                self.bins = [[0 for j in range(self.ynbins)]
                             for i in range(self.xnbins)]
            elif l.startswith("# ymin"):
                self.ymin = float(l.split('=')[1])
            elif l.startswith("# ymax"):
                self.ymax = float(l.split('=')[1])
            elif l.startswith("# out of range"):
                self.outofrange = float(l.split('=')[1])
            else:
                if len(l.split()) == 3:
                    try:
                        i, j, c = l.split()
                        self.bins[int(i)][int(j)] = float(c)
                    except Exception:
                        pass
            # end if
        # end for lines
        self.xbinwidth = (self.xmax - self.xmin) / float(self.xnbins)
        self.ybinwidth = (self.ymax - self.ymin) / float(self.ynbins)
        self.export_threshold = 0.

    def __str__(self):
        '''Dump the histogram in string format'''
        z = ''
        z += '# xnbins = {0}\n'.format(self.xnbins)
        z += '# xmin = {0}\n'.format(self.xmin)
        z += '# xmax = {0}\n'.format(self.xmax)
        z += '# ynbins = {0}\n'.format(self.ynbins)
        z += '# ymin = {0}\n'.format(self.ymin)
        z += '# ymax = {0}\n'.format(self.ymax)
        if self.export_threshold > 0:
            z += '# Exporting bins with more than {0} counts\n'.format(self.export_threshold)
        for i in range(0, self.xnbins):
            for j in range(0, self.ynbins):
                if self.bins[i][j] > self.export_threshold:
                    z += "{0} {1} {2} \n".format(i, j, self.bins[i][j])
        z += '# out of range = {0}\n'.format(self.outofrange)
        return z
    # end __str__

    def __repr__(self):
        '''Writting the histogram2D as a dictionnary'''
        z = "{"
        z += "   'xnbins':{0},\n".format(self.xnbins)
        z += "   'xmin':{0},\n".format(self.xmin)
        z += "   'xmax':{0},\n".format(self.xmax)
        z += "   'ynbins':{0},\n".format(self.ynbins)
        z += "   'ymin':{0},\n".format(self.ymin)
        z += "   'ymax':{0},\n".format(self.ymax)
        z += "   'outofrange':{0},\n".format(self.outofrange)
        z += "   'bins':{0},\n".format(str(self.bins))
        z += "}"
        return z
    # end __repr__

    def todict(self) -> dict:
        '''Return the histogram as a dict'''
        return eval(repr(self))

    def copy(self) -> 'Histogram2D':
        '''Return a copy of the histogram'''
        return Histogram2D(fromstring=str(self))

    def empty(self):
        '''Sets all bin content to 0'''
        self.outofrange = 0
        for i in range(self.xnbins):
            for j in range(self.ynbins):
                self.bins[i][j] = 0

    def empty_copy(self) -> 'Histogram2D':
        '''Returns an empty copy (i.e. preserved just the axis)'''
        return self.copy().empty()

    def rescale(self,
                newxmin: Optional[float] = None,
                newxmax: Optional[float] = None,
                newymin: Optional[float] = None,
                newymax: Optional[float] = None):
        '''Change the scale, compute new bin width, ....'''
        if newxmin is not None:
            self.xmin = newxmin
        if newxmax is not None:
            self.xmax = newxmax
        if newymin is not None:
            self.ymin = newymin
        if newymax is not None:
            self.ymax = newymax
        self.xbinwidth = (self.xmax - self.xmin) / float(self.xnbins)
        self.ybinwidth = (self.ymax - self.ymin) / float(self.ynbins)

    def dim(self) -> tuple:
        '''Alternative to len, for 2D'''
        return self.xnbins, self.ynbins

    def sum(self) -> float:
        '''compute the sum, because sum() does not work on 2D'''
        S = 0.0
        for i in range(self.xnbins):
            for j in range(self.ynbins):
                S += self.bins[i][j]
        return S

    def max(self) -> dict:
        '''returns position and value of maximum'''
        the_i, the_j, the_max = 0, 0, 0.0
        for i in range(self.xnbins):
            for j in range(self.ynbins):
                if self.bins[i][j] > the_max:
                    the_i, the_j, the_max = i, j, self.bins[i][j]
                 # end if
             # end for j
         # end for i
        return {'i': the_i, 'j': the_j, 'max': the_max}

    def scale(self, S: float = 1.):
        '''Mutliply the content by S'''
        for i in range(self.xnbins):
            for j in range(self.ynbins):
                self.bins[i][j] *= S
            # end for j
        # end for i

    def getbin(self, x: float, y: float) -> dict:
        '''return a bin information'''
        i, j = self.findbin(x, y)
        return {'i': i, 'j': j,
                'x': x, 'y': y,
                'lowedge': self.getbinlowedge(i, j),
                'updedge': self.getbinupedge(i, j),
                'count': self.bins[i][j]}

    def findbin(self, x: float, y: float) -> tuple:
        '''find a bin and returns the indices'''
        i, j = -1, -1
        if x >= self.xmin and x < self.xmax:
            i = int(floor((x - self.xmin) / self.xbinwidth))
        if y >= self.ymin and y < self.ymax:
            j = int(floor((y - self.ymin) / self.ybinwidth))
        return i, j
    # end FindBin

    def fill(self,
             x: float, y: float,
             w: float = 1.):
        '''Add one (or w) counts to bin corresponding to x

        Keyword Arguments:
        x,y -- (float)
        w -- (float) default =1'''
        i, j = self.findbin(x, y)
        if i >= 0 and j >= 0:
            try:
                self.bins[i][j] += w
            except IndexError:
                self.outofrange += w
        else:
            self.outofrange += w
    # End Fill

    def fast_fill(self,
                  x: float, y: float,
                  w: float = 1):
        '''Add one (or w) counts to the bin, in a fast way'''
        try:
            self.bins[int((x - self.xmin) / self.xbinwidth)][int((y - self.ymin) / self.ybinwidth)] += w
        except IndexError:
            self.outofrange += w

    def very_fast_fill(self,
                       i: int, j: int,
                       w: float = 1):
        ''' add one (or w) counts to the bin identified with i,j'''
        try:
            self.bins[i][j] += w
        except IndexError:
            self.outofrange += w

    def getbincenter(self, i: int, j: int) -> tuple:
        'Returns bin center'
        return self.xmin + (i + 0.5) * self.xbinwidth, self.ymin + (j + 0.5) * self.ybinwidth

    def getbinlowedge(self, i: int, j: int) -> tuple:
        'Returns bin lowedge'
        return self.xmin + i * self.xbinwidth, self.ymin + j * self.ybinwidth

    def getbinupedge(self, i: int, j: int) -> tuple:
        'Returns bin upedge'
        return self.xmin + (i + 1.) * self.xbinwidth, self.ymin + (j + 1.) * self.ybinwidth

    def xyz(self, mode: str = 'lowedge') ->tuple:
        'Returns XYZ data map for plotting'
        X = tuple((self.getbinlowedge(i, 0)[0] for i in range(self.xnbins + 1)))
        Y = tuple((self.getbinlowedge(0, j)[1] for j in range(self.ynbins + 1)))
        Z = tuple(zip(*self.bins))
        return X, Y, Z

    def project_x(self,
                  ymin: float = -1,
                  ymax: float = -1) -> Histogram:
        '''project the histogram content on the X axis'''
        jmin = self.findbin(self.xmin, ymin)[1]
        jmax = self.findbin(self.xmin, ymax)[1]
        if ymin == -1:
            jmin = 0
        if ymax == -1:
            jmax = self.ynbins
        d = [sum([self.bins[i][j] for j in range(jmin, jmax)]) for i in range(self.xnbins)]
        h = Histogram(self.xnbins, self.xmin, self.xmax)
        for i in range(self.xnbins):
            h.bins[i].count += d[i]
        return h

    def project_y(self,
                  xmin: float = -1,
                  xmax: float = -1) -> Histogram:
        '''project the histogram content on the X axis'''
        imin = self.findbin(xmin, self.ymin)[0]
        imax = self.findbin(xmax, self.ymin)[0]
        if xmin == -1:
            imin = 0
        if xmax == -1:
            imax = self.xnbins
        d = [sum([self.bins[i][j] for i in range(imin, imax)]) for j in range(self.ynbins)]
        h = Histogram(self.ynbins, self.ymin, self.ymax)
        for j in range(self.ynbins):
            h.bins[j].count += d[j]
        return h

    def rebin(self,
              newbinx: int = -1,
              newbiny: int = -1) -> 'Histogram2D':
        '''Return a new histogram with same limits but different number of bins'''
        if newbinx == -1:
            newbinx = self.xnbins
        if newbiny == -1:
            newbiny = self.ynbins
        rh = Histogram2D(newbinx,
                         self.xmin, self.xmax,
                         newbiny,
                         self.ymin, self.ymax)
        # Now filling the new histogram!
        for i in range(self.xnbins):
            for j in range(self.ynbins):
                x, y = self.getbincenter(i, j)
                rh.fill(x, y, self.bins[i][j])
        return rh

    def subhisto(self,
                 xmin: float = -1, xmax: float = -1,
                 ymin: float = -1, ymax: float = -1) -> 'Histogram2D':
        '''Return a new 2D histogram, with only the selected range'''
        # First, select
        jmin = self.findbin(self.xmin, ymin)[1]
        jmax = self.findbin(self.xmin, ymax)[1]
        imin = self.findbin(xmin, self.ymin)[0]
        imax = self.findbin(xmax, self.ymin)[0]
        if xmin == -1:
            imin = 0
        if xmax == -1:
            imax = self.xnbins
        if ymin == -1:
            jmin = 0
        if ymax == -1:
            jmax = self.ynbins
        jmin, jmax = min(jmin, jmax), max(jmin, jmax)
        imin, imax = min(imin, imax), max(imin, imax)
        subh = Histogram2D(imax-imin,
                           self.getbinlowedge(imin, jmin)[0],
                           self.getbinlowedge(imax, jmin)[0],
                           jmax - jmin,
                           self.getbinlowedge(imin, jmin)[1],
                           self.getbinlowedge(imin, jmax)[1])
        subh.bins = [[self.bins[i][j] for j in range(jmin, jmax)]
                     for i in range(imin, imax)]
        return subh

    def autocrop(self, threshold: float = 0):
        '''Cut the bins that are 'empty' '''
        # Find the first i,j that is above threshold

        # use specialized function:
        def _ifromthetop():
            for i in range(self.xnbins):
                for j in range(self.ynbins):
                    if self.bins[i][j] > threshold:
                        return i
            return -1

        def _jfromthetop():
            for j in range(self.ynbins):
                for i in range(self.xnbins):
                    if self.bins[i][j] > threshold:
                        return j
            return -1

        def _ifromthebottom():
            for i in range(self.xnbins - 1, 0, -1):
                for j in range(self.ynbins - 1, 0, -1):
                    if self.bins[i][j] > threshold:
                        return i
            return 0

        def _jfromthebottom():
            for j in range(self.ynbins - 1, 0, -1):
                for i in range(self.xnbins - 1, 0, -1):
                    if self.bins[i][j] > threshold:
                        return j
            return 0

        i_top, j_top = _ifromthetop(), _jfromthetop()
        i_bot, j_bot = _ifromthebottom(), _jfromthebottom()
        i_top, i_bot = max(i_top, i_bot), min(i_top, i_bot)
        j_top, j_bot = max(j_top, j_bot), min(j_top, j_bot)
        xmin, ymin = self.getbinlowedge(i_bot, j_bot)
        xmax, ymax = self.getbinupedge(i_top, j_top)
        return self.subhisto(xmin, xmax, ymin, ymax)
# end of class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Oct  9 11:31:54 CEST 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: lazy_histogram.py -*-
# -*- purpose: -*-

'''
Ghost Histogram has the outside API of a real one,
but nothing happens behind.
This is to be used in filling situation, where it's easier to
fill a 'ghost' rather than test if the channel is defined, ...
'''


class GhostHistogram:
    '''
    Behave like a 1D histogram, but does nothing
    '''
    def __init__(self, *w, **kw):
        self.nbins = 0
        self.xmin = 0
        self.xmax = 0
        self.binwidth = 0.
        self.bins = []
        self.outofrange = 0

    def __str__(self):
        '''Dump the histogram in string format'''
        z = f'# nbins = {self.nbins}\n'
        z += f'# xmin = {self.xmin}\n'
        z += f'# xmax = {self.xmax}\n'
        for i in range(0, self.nbins):
            lowedge = self.xmin + i * self.binwidth
            z += f"{lowedge} {self.binwidth} {self.bins[i]}\n"
        z += f'# out of range = {self.outofrange}\n'
        return z
    # end __str__

    # Next, define cContainer like behavior
    def __getitem__(self, k):
        return self.bins[k]

    def __setitem__(self, k, v):
        self.bins[k] = v

    def __len__(self) -> int:
        return len(self.nbins)

    def fast_index(self, x) -> int:
        ' false index finder '
        return 0

    def __iter__(self):
        ''' provides iteration over bin directly '''
        return iter(self.bins)

    def fast_fill(self, x: float,
                  w: int = 1):
        ' false filler '
        pass
# end of class

# EOF



#!/usr/env python
# -*- coding: utf-8 -*-
# -*- format: python -*-
# -*- author: G. Henning -*-
# -*- created: Tue Dec  4 10:59:10 CET 2018 -*-
# -*- copyright: GH/IPHC 2018 -*-
# -*- file: parzen_estimator.py -*-
# -*- purpose: -*-

'''
Made o estimate the value of the function follwing the histogram
(not fully implemented yet)
'''

import math

import pyhisto
from pyhisto import Histogram


def kernel_gauss(u, w) -> float:
    '''Gaussian kernel'''
    return 0.39894228 * math.exp(-.5 * u * u)

def kernel_parabol(u, w) -> float:
    '''Parabolic kernel'''
    if (u >=- 1 and u <= 1):
        return 3. / 4. * ( 1. - u * u)
    return 0.0


class ParzenEstimator:
    '''Create an estimator'''

    def __init__(self,
                 h,
                 kernel=kernel_gauss,
                 width=lambda b: b.width(),
                 normalize=True):
        raise NotImplementedError
        self.h = h.copy()
        self.width = width
        self.kernel = kernel

        def estim(x):
            ' internal function for estimation '
            f = 0.0
            for b in h:
                w = self.width(b)
                u = (x - b.center()) / w
                f+= self.kernel(u, w) * b.count
            f /= 1.0 * len(self.h)
            return f
        self.Norm = 1.0
        if normalize:
            S = sum(self.h)
            estS = 0.0
            nint = int(self.h.nbins * 2.33)
            wint = (self.h.xmax - self.h.xmin) / nint
            for i in range(nint):
                estS += estim(self.h.xmin + i * wint ) * wint
            self.Norm = S / estS
        elif normalize == 'to1':
            estS = 0.0
            nint = int(self.h.nbins * 2.33)
            wint = (self.h.xmax - self.h.xmin) / nint
            for i in range(nint):
                estS += estim(self.h.xmin + i * wint) * wint
            self.Norm = 1. / estS
        self.estimator = estim

    def __repr__(self):
        return "<ParzenEstimator of {0}>".format(repr(self.h))

    def __call__(self, x):
        return self.estimator(x) * self.Norm

# end of class

# EOF

